use std::collections::HashMap;

use mlua::IntoLua;
use serde::{Deserialize, Serialize};

/// Converts a `serde_json::Value` to a `mlua::Value`.
fn json_to_lua(value: serde_json::Value, lua: &mlua::Lua) -> mlua::Result<mlua::Value<'_>> {
    match value {
        serde_json::Value::Null => Ok(mlua::Value::Nil),
        serde_json::Value::Bool(value) => Ok(mlua::Value::Boolean(value)),
        serde_json::Value::Number(value) => match value.as_f64() {
            Some(number) => Ok(mlua::Value::Number(number)),
            None => Err(mlua::Error::ToLuaConversionError {
                from: "serde_json::Value::Number",
                to: "Number",
                message: Some("Number cannot be represented by a floating-point type".into()),
            }),
        },
        serde_json::Value::String(value) => lua.create_string(value).map(mlua::Value::String),
        serde_json::Value::Array(value) => {
            let tbl = lua.create_table_with_capacity(value.len(), 0)?;
            for v in value {
                let v = json_to_lua(v, lua)?;
                tbl.push(v)?;
            }
            Ok(mlua::Value::Table(tbl))
        }
        serde_json::Value::Object(value) => {
            let tbl = lua.create_table_with_capacity(0, value.len())?;
            for (k, v) in value {
                let k = lua.create_string(k)?;
                let v = json_to_lua(v, lua)?;
                tbl.set(k, v)?;
            }
            Ok(mlua::Value::Table(tbl))
        }
    }
}

#[derive(Clone, Debug, Serialize)]
pub(crate) struct WebhookEvent {
    pub topic: String,
    pub query: HashMap<String, String>,
    pub body: serde_json::Value,
}

#[derive(Clone, Debug, Serialize)]
pub(crate) struct ApiEvent {
    pub id: String,
    pub body: serde_json::Value,
    pub headers: HashMap<String, Vec<u8>>,
    pub status: u16,
}

impl<'lua> IntoLua<'lua> for ApiEvent {
    fn into_lua(self, lua: &'lua mlua::Lua) -> mlua::Result<mlua::Value<'lua>> {
        let headers = lua.create_table_with_capacity(0, self.headers.len())?;
        for (k, v) in self.headers {
            let k = lua.create_string(k)?;
            let v = lua.create_string(v)?;
            headers.set(k, v)?;
        }

        let body = json_to_lua(self.body, lua)?;
        let tbl = lua.create_table_with_capacity(0, 4)?;

        lua.create_string(self.id)
            .and_then(|id| tbl.set("id", id))?;
        tbl.set("status", self.status)?;
        tbl.set("headers", headers)?;
        tbl.set("body", body)?;

        Ok(mlua::Value::Table(tbl))
    }
}

#[derive(Clone, Debug, Serialize)]
pub(crate) struct ErrorEvent {
    pub id: String,
    pub message: String,
}

#[derive(Clone, Debug)]
pub(crate) enum Event {
    Webhook(WebhookEvent),
    Api(ApiEvent),
    Error(ErrorEvent),
}

impl Event {
    pub fn error(id: String, message: impl ToString) -> Self {
        Self::Error(ErrorEvent {
            id,
            message: message.to_string(),
        })
    }
}

#[derive(Clone, Debug, serde_repr::Deserialize_repr, serde_repr::Serialize_repr)]
#[repr(u8)]
pub(crate) enum NtfyPriority {
    Min = 1,
    Low = 2,
    Default = 3,
    High = 4,
    Max = 5,
}

impl Default for NtfyPriority {
    fn default() -> Self {
        Self::Default
    }
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub(crate) struct NtfyAction {
    pub action: String,
    pub label: String,
    pub url: String,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub(crate) struct NtfyMessage {
    // Needs to be deserialized from Lua, but should not be sent to Ntfy
    #[serde(skip_serializing)]
    pub url: String,
    // Needs to be deserialized from Lua, but should not be sent to Ntfy
    #[serde(skip_serializing)]
    pub token: String,
    pub topic: String,
    pub message: String,
    #[serde(default)]
    pub title: Option<String>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub tags: Vec<String>,
    #[serde(default)]
    pub priority: NtfyPriority,
    #[serde(default)]
    pub click: Option<String>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub actions: Vec<NtfyAction>,
    #[serde(default)]
    pub markdown: bool,
    #[serde(default)]
    pub icon: Option<String>,
    #[serde(default)]
    pub delay: Option<String>,
}

#[derive(Clone, Debug)]
pub(crate) enum Message {
    Ntfy(NtfyMessage),
}

#[derive(Clone, Debug, Deserialize)]
pub(crate) enum Method {
    #[serde(alias = "GET", alias = "get")]
    Get,
    #[serde(alias = "POST", alias = "post")]
    Post,
}

impl Default for Method {
    fn default() -> Self {
        Self::Get
    }
}

#[derive(Clone, Debug, Deserialize)]
pub(crate) struct ApiTask {
    pub id: String,
    #[serde(default)]
    pub delay: u64,
    #[serde(default)]
    pub method: Method,
    pub url: String,
    #[serde(default)]
    pub body: serde_json::Value,
    #[serde(default)]
    pub query: HashMap<String, String>,
    #[serde(default)]
    pub headers: HashMap<String, String>,
}
